<?php

 define('iODLeft','{#');
 define('iODRight','#}');
 define('iOR','||');
 define('iDelimLeft','{$');
 define('iDelimRight','$}');
 
 /*
 
  Simple macro:
  
  SELECT * 
  FROM tblName AS t
  WHERE t.fldName='{$fldName$}'
  
  {$fldName$} replaced with value of activeTable.fldName
  
  Simple macro has potential problem. If you marked tblName in relation with current active table activeTable,
  and activeTable does not have fldName in it, macro will be unable to execute, and vocabulary will
  return empty result. It is wrong way, so that it is strongly recommended to use Smart macro,
  which is able to replace whole query part, dependent on macro, with valid, meaningful placeholder,
  in order to make sure, that you get query working properly anyway.
  
  In other words, everything found between {# and #} will be replaced by the part, found between || and #}
  in cases, when macro can not find apropriate value for substitution.
  
  ex: {#aaaaa{$bbb$}ccc||ddd#} => aaaaaVALUEccc OR ddd
  ex: {#aaaaa{$bbb$}ccc{$dddd$}eeeee||fff#} => aaaaaVALUE1cccVALUE2eeee OR fff
  
  Smart macro:
  
  SELECT * 
  FROM tblName AS t
  {#WHERE t.fldName='{$fldName$}'||somePlaceHolderCode#}
  
  {$fldName$} replaced with value of activeTable.fldName, if it is present in current variables set.
  In other case it is replaced with somePlaceHolderCode as it is provided. Usually you will want 
  somePlaceHolderCode be either '1=1', or '' (empty space).
  
  ex: 
  
  SELECT * 
  FROM tblName AS t
  WHERE {#t.fldName='{$fldName$}'||1=1#}
  
  if activeTable.fldName does not exist, th request will be converted to:
  
  SELECT * 
  FROM tblName AS t
  WHERE 1=1
  
  actually, if you don't have complex WHERE in your request, it's better to remove WHERE ... completely.

  ex:
  
  SELECT * 
  FROM tblName AS t
  {#WHERE t.fldName='{$fldName$}'||#}

  if activeTable.fldName does not exist, th request will be converted to:
  
  SELECT * 
  FROM tblName AS t

  ex:
  
  SELECT * 
  FROM tblName AS t
  WHERE {#t.fldName='{$fldName$}'||1=1#} AND t.flag=1
  
  if activeTable.fldName does not exist, th request will be converted to:
  
  SELECT * 
  FROM tblName AS t
  WHERE 1=1 AND t.flag=1
  
 */

 function varExport($var) {
  
  $e = print_r($var,TRUE);
  file_put_contents('z:\tmp\print_r.txt',$e);
 }
 
 
 
 
 

 function iExtractMacros($query="") {
  $ret = array();
  /*
   Check if we have at least one macro inside the $query
  */
  if (strpos($query,iDelimLeft)!==FALSE && strpos($query,iDelimRight)!==FALSE) {
   /*
    Cut $query into pieces, containing macro+rest of the string
   */
   $tmp = explode(iDelimLeft,$query);
   $l=count($tmp);
   /*
    Ignore first piece, because it doesn't contain any macros
   */
   for($i=1;$i<$l;$i++) {
    /*
     Cut the rest of the string from macro
    */
    $s = trim(substr($tmp[$i],0,strpos($tmp[$i],iDelimRight)));
    /*
     If proper macro, like {$fieldName$} found, save it into array $ret
    */
    if (strlen($s)>0) $ret[] = $s;
   }
  }
  /*
   return found macros in the order they've been met
  */
  return $ret;
 }

 
 
 function iSmartExtractMacros($query="") {
  $ret = array();
  /*
   Check if we have at least one macro inside the $query
  */
  if (strpos($query,iODLeft)!==FALSE && strpos($query,iODRight)!==FALSE) {
   /*
    Cut $query into pieces, containing macro+rest of the string
   */
   $tmp = explode(iODLeft,$query);
   $l=count($tmp);
   /*
    Ignore first piece, because it doesn't contain any macros
   */
   for($i=1;$i<$l;$i++) {
    /*
     Cut the rest of the string from macro
    */
    $ss = trim(substr($tmp[$i],0,strpos($tmp[$i],iODRight)));
    $s = $ss;
    $avp = strpos($s,iOR);
    if ($avp===FALSE) { //Alternative value was not provided, take default value for it
     $av = '';
    } else {
     $av = substr($s,$avp+strlen(iOR));
     $s = trim(substr($s,0,$avp));
    }
    /*
     If proper macro, like {$fieldName$} found, save it into array $ret
    */
    if (strlen($s)>0) $ret[] = array('query'=>$s, 'aValue'=>$av, 'source'=>$ss);
   }
  }
  /*
   return found macros in the order they've been met
  */
  return $ret;
 }

 
 
 function iReplaceMacros($query="",$values=array(),$tablename="",$vlname="") {
  /*
   $values = $this->app->currentRecord->_values[];
  */

  /*
   extract any macro like {$fieldName$} from the __sql__ query, 
   given in valuelists.ini files.
  */
	$marr = iExtractMacros($query);
	if ($marr) {
		foreach($marr as $macro) {
		  /*
		   replace each macro with the corresponding field value, if it exists.
		   if it doesn't exist, macro will not be replaced, so that query,
		   if executed, will return wrong result.

		   Because of this we should raise an exception anytime, we found
		   macro for undefined fieldName.
      */
			if (isset($values[$macro])) {
			  /*
			    If requested in $macro fieldName exists, we replace macro within
			    fieldName's value
        */
				$query = str_replace(iDelimLeft.$macro.iDelimRight,$values[$macro],$query);
			} else {
			  /*
			    If macro tries to operate with undefined fieldName, 
			    we return FALSE immediately.
              */
             return FALSE;
			}
		}
	}
	return $query;
 }
 
 
 
 function iSmartReplaceMacros($query="",$values=array(),$tablename="",$vlname="") {
  /*
   $values = $this->app->currentRecord->_values[];
  */

  /*
   extract any macro like {#aaa{$fieldName$}bbb||ccc#} from the __sql__ query, 
   given in valuelists.ini files.
  */
	$smarr = iSmartExtractMacros($query);
  
  //echo nl2br(print_r($smarr,TRUE));
  
	if ($smarr) {
		foreach($smarr as $smacro) {
		  /*
		   replace each macro with the corresponding field value, if it exists.
		   if it doesn't exist, macro will not be replaced, so that query,
		   if executed, will return wrong result.

		   Because of this we should raise an exception anytime, we found
		   macro for undefined fieldName.
      */
			if ($m=iReplaceMacros($smacro['query'],$values,$tablename,$vlname)) {
			  /*
			    If requested in $macro fieldName exists, we replace macro within
			    fieldName's value
        */
				$query = str_replace(iODLeft.$smacro['source'].iODRight,$m,$query);
			} else {
			  /*
			    If macro tries to operate with undefined fieldName, 
			    we replace whole macro with alternative value or empty space.
        */
				$query = str_replace(iODLeft.$smacro['source'].iODRight,$smacro['aValue'],$query);
			}
		}
	} else {
    /*
     No smart macro provided, let's try to detect simple macros...
    */
     if (strlen(trim($query))>0 && !($q = iReplaceMacros($query,$values,$tablename,$vlname))) {
      /* 
       There was at least unsuccessfull macro. Actually we can not continue with it
       and should throw exception, because if we don't, there would be problems within
       table relations and data substitution. So, use Smart Macros, to correct this situation.
      */
      
      throw new Exception('Use Smart macro instead Simple macro in __sql__ => [./appName/tables/'.$tablename.'/valuelists.ini] => ['.$vlname.'] => query: ['.$query.']');
      
     }
     
     $query = $q;
    }
    
	return $query;
 }
?>